rm(list=ls())

###############################################################
####
####  Copy Number recognition in Affymetrix SNP 6.0 arrays
####
####   This procedure takes segm_Sample1 files, applies GAP parameters from GAP_parameters_file,
####     and makes estimation of segmental Copy number and Genotype.
####    Results are saved in Results_Plots folder as: 
####      (1) proc_Sample1 text file of all segments and copy numbers
####      (2) Chr_Plot_Sample1 per chromosome profiles .jpg
####      (3) Summary_Plot_Sample1 summary profiles and information
####      (4) GAP_Plot_Sample1 plot of GAP and copy number detection
####
##################################################################

### !!!!! Put the PATH to your data  !!!!

pPath<-"C:/GAP_AffymetrixSNP6_workflow_homepage/"

inputPath<-paste(pPath,"Vectors_and_Segmentations/",sep="")
outputPath<-paste(pPath,"Results_Plots/",sep="")
outputPath_Plot<-paste(pPath,"Results_Plots/",sep="")


### Annotation files
### !!!!! Correct the names of annotation files according to script_0  !!!!

CN_annot<-read.csv(paste(pPath,"CN_annot_Affy6_Version.csv",sep=""),header=T,fill=T,stringsAsFactors = F,row.names=1)
SNP_annot<-read.csv(paste(pPath,"SNP_annot_Affy6_Version.csv",sep=""),header=T,fill=T,stringsAsFactors = F,row.names=1)

CN_annot<-data.matrix(CN_annot)
SNP_annot<-data.matrix(SNP_annot)
C_Ind<-1;C_Chr<-2;C_Pos<-4

### Both source files need SNP_annot file defined above
source(paste(pPath,"script_functions_GAP_recognition_Affy6.R",sep=""))
source(paste(pPath,"script_functions_GAP_recognition_Affy6_fit.R",sep=""))


### !!!!! If you want, put some other file name for summary file  !!!!
output_infoFileName<-"GAPrecognitionInfo_Affy6.txt"


### !!!!! Check the GAP_parameters_file_name, te be the same as after script_3   !!!!
GAP_parameters_file_name<-"GAP_parameters_Example.csv"


if(file.access(paste(inputPath,GAP_parameters_file_name,sep=""), mode = 0)==0){
	GAP_par_exist<-T
	GAP_parameters_file<-read.csv(paste(inputPath,GAP_parameters_file_name,sep=""),stringsAsFactors = F,header=T,row.names=1)
	GAP_parameters_file<-data.matrix(GAP_parameters_file)
}else{GAP_par_exist<-F}


######## Structure of output summary information file concerning recognition
######## and rearrangements for all samples in the list
info<-c("p_BAF","q_LRR","2copyLRR","Breaks","DNAindex","Ploidy_detected","Chromosome_counts")
cat(c("SampleID",info), file = paste(outputPath,output_infoFileName,sep=""),sep="\t","\n")


#### !!!!! All segm_Sample1 files from Vectors_and_Segmentations folder will be processed !!!!

segFiles <- list.files(inputPath, pattern=vect_Segm_pattern,full.names=T) 

fileNames <- gsub(inputPath,"",segFiles) 
fileNames <- gsub(vect_Segm_pattern,"",fileNames) 
fileNames <- gsub(".txt","",fileNames) 

fileNames

nSample<-length(segFiles)

##########################   MAIN CYCLE  ######################################

ii<-1

for(ii in 1:nSample){


print(paste(as.character(ii),"out of",as.character(nSample),"processing",fileNames[ii],sep="  "))


tmp<-read.delim(segFiles[ii],header=T,sep="\t")
plot(tmp[,c_adL],tmp[,c_cn],cex=tmp[,c_len]/quantile(tmp[,c_len],prob=0.9),
		main="GAP",xlab="Allelic Difference",ylab="Log ratio")
abline(v=0)
abline(-1,1)


#################################################################
####
#### Step 1.  Determination of GAP pattern
####  		functions: CopyNumber_and_Genotype_pattern
####               and  CopyNumber_and_Genotype_patternWOfitting
####
##################################################################

print(paste("   fitting GAP  ",fileNames[ii]))
notNAs<-intersect(which(!is.na(tmp[,c_cn]+tmp[,c_adL])),which(tmp[,c_chr]<23))

############### Comments to notNAs
##### NAs are not allowed in recognition function
##### Sex chromosomes on SNP arrays are usually of rather low quality (eg, higher variance)
##### Y chromosome in female sample gives oulying values, so
##### it might be a reason to exclude sex chromosomes from pattern recognition procedure


###################################################################
###
### Three ways of dealing with GAP pattern recognition function: 
###
### 1. If GAP parameters were precalculated, use function CopyNumber_and_Genotype_patternWOfitting
###  The function produce all necessary things for recognition with the given constants
###
###  2. Use this function also if automatical recognition fails to fit the model correctly, while
###   one can see the right solution by visual inspection of the GAP pattern. 
###   Introduce corresponding values in the model and look at resulting pattern
###   Adjust parameters if necessary
###
###  3. If performing recognition de novo use function CopyNumber_and_Genotype_pattern


##### Choose parameters from GAP_parameters_file

GAP_parameters<-F

if(GAP_par_exist){
ii_name<-grep(fileNames[ii],rownames(GAP_parameters_file))
if(length(ii_name)!=1){
if(length(ii_name)>1){
	ii_name1<-numeric(0)
	for(i in 1:length(ii_name)){ii_name1<-c(ii_name1,grep(rownames(GAP_parameters_file)[ii_name[i]],fileNames[ii]))}
	if(length(ii_name1)==1){ii_name<-ii_name[ii_name1];GAP_parameters<-T}
}
}else{GAP_parameters<-T}
}


if(GAP_parameters){

### If precalculated GAP parameters: case 1.

p_BAF<-GAP_parameters_file[ii_name,2];q_LRR<-GAP_parameters_file[ii_name,3];Delta<-c(2,GAP_parameters_file[ii_name,4])
result<-CopyNumber_and_Genotype_patternWOfitting(tmp[notNAs,c_adL],tmp[notNAs,c_cn],tmp[notNAs,c_len],showGAPTemplate=T,power=3/4,p_BAF,q_LRR,Delta)

}else{

### If calculating GAP parameters de novo: case 3.

result<-CopyNumber_and_Genotype_pattern(tmp[notNAs,c_adL],tmp[notNAs,c_cn],tmp[notNAs,c_len],showGAPTemplate=T,power=3/4)

}

#### My parameters: case 2.
#### Delta is the vector of 2: c(copy number, LR value)
#### For example, sample BLC_T31 automatically detected as diploid, but, it looks like just dublicated diploid (near-tetraploid)
#### so, we introduce: p_BAF<-0.7;q_LRR<-0.13;Delta<-c(2,-0.06)
#### result<-CopyNumber_and_Genotype_patternWOfitting(tmp[notNAs,c_adL],tmp[notNAs,c_cn],tmp[notNAs,c_len],showGAPTemplate=T,power=3/4,p_BAF,q_LRR,Delta)

### coefficients of the pattern shrinkage
p_BAF<-result$coeff[1]
q_LRR<-result$coeff[2]

### matrix of coordinates of claster centers
Centr_CL<-result$Centr_CL

### matrix of clusters on the LRR x BAF plane
NETM<-result$NETM

### matrix of coordinates of subclones
SB_CL<-result$SB_CL

rm(result)

############# GAP pattern recognition is completed!

###################### Comments to GAP recognition results
#### As a result we have got: 
####		1) p_BAF and q_LRR: parameters of the model GAP
####		2) Centr_CL:  a table with coordinates of model GAP clusters
####		3) NETM: a matrix enumerating all grid nodes and their cluster annotation 
####			(Index in columname of NETM = Index in rownames of Centr_CL)
####		4) SB_CL: a matrix enumerating all segments positioned essentially inbetween clusters
####			(Index_Subclone in rowname of SB_CL = Index in columname of NETM)
#### These data are used for copy number and genotype attribution to each segment in tmp table


##########################################################
############## Output segmented summary file format with Copy Number, Major allele and other supplemetaries

tmp<-tmp[,1:12]
tmp<-cbind(tmp,matrix(0,dim(tmp)[[1]],16))
colnames(tmp)[13:28]<-ColNamesTMP[13:28]


#################################################################
####
#### Step 2.  Assigning Copy number and Genotype to each segment in tmp table
####	 based on the closest cluster annotation ("closest" in terms of maximal likelyhood of AD distribution)
####	    + iterative fitting of small segments to big segments if possible             
####
##################################################################


######################### Comments(1) to CN recognition
### function getCNwithPrecision determines the closest Copy number from the GAP pattern to each smoothed Log Ratio value
### Segments within Precision relative distance around CN level are attributed pure CN status
### Segments beyond Precision relative distance around CN level are attributed rational CN 
### (such as 2.3, meaning at 3/10 to 2 copy and 7/10 to 3 copy level)
##### Further, CN and GT attribution is performed based on max LogLikelyhood cluster
##### (segments with 2.3 CN status are checked to have AA, AB, AAA, AAB genotypes: maxLike one will be chosen)
##### (segments with 2 CN status are checked to have AA, AB genotypes: maxLike one will be chosen)

################## Comments (2) to recognition results
##### If Allelic differences are calculated using the function getADforGAPrecogn,
#####   being suggested in the current sources, the results need some additional processing
##### The reason for this: function getADforGAPrecogn was constructed to extract ONE of the two modes
#####  of allelic difference distribution, maximally avoiding FALSE values.
##### Thus, homozygous mode was often used instead of heterozygous.
##### This provides a number of attributions such as those with 0 in B allele, meaning germline Homozygosity
##### One can see many 0 in B allele for almost pure tumors with p_BAF < 0.15
##### In that case it is hard to distinguish germline and accuired homozygosity
##### Various heuristics could be used in these cases
#####
#####   (0 in B alelle means germline homozygosity with undeterminable somatic allelic status)


print(paste("   adjusting CN profile  ",fileNames[ii]))

tmp<-getCNwithPrecision(tmp,Centr_CL,Precision=1/10)
Break<-Centr_CL[5,21]-Centr_CL[5,20]
tmp<-trueCNbreakCount_new(tmp,Break)
tmp<-fitCNbreakCount_short(tmp)


##### For this procedure the vector of Allelic differences is necessary
datA<-read.table(sub(vect_Segm_pattern,vect_AD_pattern,segFiles[ii]),header=F)
datA<-data.matrix(datA)
datA<-as.numeric(datA)
datA<-abs(datA)

##### Estimating Standard deviation of homozygous band of Allelic Differences
sdH<-getSDofHomozygousADband(tmp,abs(datA),Length=100,centerBoundary=0.3)

##### Estimating Copy number and Allelic counts based on log likelyhood 
tmp<-CNandGTbyLogLike_withSD_Rank(tmp,abs(datA),Centr_CL,sdH,Tolerance=0.1)

tmp[,c_CN]<-tmp[,c_CN1]
tmp[,c_BA]<-tmp[,c_BA1]

################ Comments (3) to the function CNandGTbyLogLike_withSD_Rank
##
## function CNandGTbyLogLike_withSD_Rank calculates max LogLikelyhood of 
##  Alleic difference 2 mode distribution for all clusters determined by GAP,
## as a mixture of two normals with fixed Means fitting only SDs and Lamdas.
## If proportion of heterozygous mode is less than 0.1, segment is reported to be 
## homozygous with corresponding CN.
##
## Results are placed into columns CN1,BA1,rnk1, for the first choice, CN2,BA2,rnk2 for the second, etc 
## rnk1,2,3 are set automatically to 1,2 and 3 (rnk are ranks of attribution).
## Ranks are adjusted to be 1,1,3 (or 1,1,1, etc), if indistinguishable with respect to Tolerance LogLikes.
## CN2(3) are set to -1, if CN2(3) has significantly lower logLike
## Tolerance specifies which LogLikes are considered to be different
## True and False solutions often have very close LogLikes
## To keep proper ranking we have included Tolerance

############## Comments (4) to recognition procedure
#####
##### If the sample is noisy and highly contaminated, and segmentation has a lot of false breaks,
##### one can use smoothing procedure which is based on recognition pattern.
##### The basic principles are: take a big well recognized regions as a true one 
##### and fit, if possible, relevant small segments with ambiguos attribution
#####


##### Doublechecking the segments for balanced genotype (AB, AABB, AAABBB, etc)
tmp[,c_CN1:(c_CN1+8)]<-abs(tmp[,c_CN1:(c_CN1+8)])
tmp<-getADconfForBalanced(tmp,datA,Centr_CL)

##################### Comments (5) to Doublechecking..
####    The problem for highly contaminated samples
####    is to distinguish balanced and not balanced AD bands as they are very close
####    and often show similar logLikes


########### Fitting procedure
tmp<-iterativeSegmentsFit(tmp,segLength=c(1000,800,400,200,100))

############## Comments (6) to Fitting procedure
####   Segments are iterativelly fitted, if possible, to stable neihbouring segments
####   Iterations are performed according to the length of the segments
####   We used 1000 SNPs as a basis segment length, going further to 800,400,200,100 SNPs
####   Other length combinations were not tested. 
####  As a result CN and GT are ascribed to the segments: see tmp columns Copy and MAllele
####

#### Fitting short (<100 SNPs) segment to surrounding solid ones, if possible
tmp<-shortBreaksToTMP(tmp)

############## Comments (7) to Fitting short segments
#### Short segments in CN profile are often the results of 
####  redundant CBS segmentation of experimental noise.
#### Function is checking, if the short segment really corresponds to CN or AD alteration


####### Final polish 

tmp<-iliminateShort_FinalPolish(tmp)

tmp<-checkSmallHomozygosities_FinalPolish(tmp,Length=399)
tmp<-checkHomozygosities_FinalPolish(tmp,Length=400,p_BAF)
tmp<-checkHomozygBreaks_finalPolish(tmp)
tmp<-checkSmallDiscrepancies_finalPolish(tmp,Length=200)
tmp<-checkSmallHomozygosities_FinalPolish(tmp,Length=399)

tmp<-getConfidence(tmp)

############## Comments to Final Polish
#### Adjusting changes in  allele counts for high copy number segments,
#### fitting short germline homozygosities 
#### and eliminating segments less than 10 SNPs.


############# OUTPUT ##################################

###### Smoothing and saving the result

print(paste("   saving results  ",fileNames[ii]))

datC<-read.table(sub(vect_Segm_pattern,vect_CN_pattern,segFiles[ii]),header=F)
datC<-data.matrix(datC)
datC<-as.numeric(datC)

CHR<-CHR_table;

tmp<-updateMeans_FinalPolish(tmp,datA,datC,Centr_CL,p_BAF)


###### Add chromosomal position to segmented table

tmp<-cbind(tmp[,1:3],tmp[,2:dim(tmp)[[2]]])
tmp[,4]<-SNP_annot[tmp[,c_is],C_Pos]
tmp[,5]<-SNP_annot[tmp[,c_if],C_Pos]
colnames(tmp)[4:5]<-c("PosSNPstart","PosSNPend")

fName<-paste(outputPath,"proc_",fileNames[ii],".txt",sep="")
write.table(tmp, file = fName,quote=F,row.names=F,sep="\t")

tmp<-tmp[,-c(4,5)]


####### Summary figures

tmp<-shrinkReprTMP(tmp)

GAPplot(tmp,nameSample=fileNames[ii],Centr_CL,NETM,q_LRR)

summaryPlots_perChrom(tmp,nameSample=fileNames[ii], datA, datC, Centr_CL)

info<-summaryPlot(tmp,nameSample=fileNames[ii], datA, datC, Centr_CL,NETM,q_LRR)

cat(c(fileNames[ii],info), file = paste(outputPath,output_infoFileName,sep=""),sep="\t",append=T,"\n")

}##ii

############  END MAIN CYCLE  #############################################
#dev.off()
